/*
  CPP_CONTEST=2013
  CPP_PROBLEM=D
  CPP_LANG=C+OPENMP+MPI
  CPP_PROCESSES_PER_NODE=marte 1 mercurio 1
*/

//

/* RECORD
José Manuel Cruz Zapata 
student of the University of Murcia, Spain
records established December 10th, 2013
time 1139 msec
speed-up 11.85

reevaluated in September 10th, 2014
time 1271
speed-up 10.62
*/ 

//
/*Uses one process in each node
and as many threads inside each process as number of cores in the node (6).
Additional speed-up is achieved with sequential optimization*/

#include <stdlib.h>
#include <omp.h>
#include <mpi.h>
 
double onetaylor(int N,double *a,double x,double y)
{
  int i,j;
  double r=0,v,v2;

  // No utilizamos matriz ni vector sino que vamos calculando sobre la marcha, siendo necesarias únicamente dos
  // variables para guardar los valores que necesitamos para los nuevos cálculos

  // v acumula el valor anterior para calcular cada fila
  // v2 acumula el valor anterior para empezar el cálculo de una nueva fila

  // Inicializamos v y v2:
  v = 1;
  v2 = 1;  

  // Cálculo de cada posición partiendo de la posición anterior.
  // Al movernos por columnas dividimos entre y, por filas entre x.
  for(i=0;i<N;i++) { 

    // Primera posición se trata distinta (no se divide v entre y) 
    r+=v*a[i*N];

    // Resto de posiciones de la fila:   
    for(j=1;j<N-i;j++) { 
      v = v/y;     
      r+=v*a[i*N+j];      
    }

    v2 = v2/x;
    v = v2;
  }  

  // Devolvemos el valor calculado:
  return r;
}
    
void  sec(int N,double *a,int P,double *x,double *y,double *r,int nodo,int np)
{
  int i, position, bloque, sobrante, indice, bloqueEnviado;
  MPI_Status status;

  omp_set_num_threads(6);

  // Repartimos N y P entre todos los procesos
  MPI_Bcast(&N,1,MPI_INT,0,MPI_COMM_WORLD);
  MPI_Bcast(&P,1,MPI_INT,0,MPI_COMM_WORLD);
  // Repartimos a
  if (nodo != 0)  a = (double *) malloc(sizeof(double)*N*N);
  MPI_Bcast(a,N*N,MPI_DOUBLE,0,MPI_COMM_WORLD);

  /* 
     Para repartir los vectores entre los procesos haremos una especie de scatter manual, dividiendo dichos vectores
     en bloques que se asignarán a cada proceso. La razón de hacerlo de forma manual es poder empaquetar los bloques
     de cada vector de forma que sólo se realice un envío a cada uno de los procesos.
  */

  /*
	Dividimos x e y entre los procesos de la siguiente forma:
	Se toma la división redondeada hacia abajo y se reparten las posiciones sobrantes empezando por el primer proceso.
	De esta forma, por ejemplo, con P*P=9 y 4 procesos tendremos inicialmente bloques de 2 posiciones repartidas entre
	cada uno de los cuatro procesos y la columna restante se asigna al proceso 0.
  */
  bloque = (P*P)/np; // Division entera redondeada hacia abajo
  sobrante = (P*P)%np; // Iteraciones de P que restan tras dividir P exactamente entre np

  // Tamaño máximo del paquete que se enviará a cada proceso (1 bloque de cada uno de los 2 vectores)
  int tamPack = (bloque+1)*sizeof(double)*2;
  char buffer[tamPack];

  if (nodo == 0) {
	// Para cada uno de los nodos:
	for (i=1; i<np; i++) {
		// Vemos si le corresponde el tamaño mínimo de bloque o tiene alguna columna extra
		if (i < sobrante) bloqueEnviado = bloque+1;
		else bloqueEnviado = bloque;

		position = 0;

		// Calculamos el índice teniendo en cuenta que puede haber bloques de dos tamaños distintos en
		// función de como divida np a P
		if (i<=sobrante) indice = i*(bloque+1);
		else indice = sobrante*(bloque+1) + (i-sobrante)*bloque;

		// Empaquetamos sus trozos de los vectores, copiando bloque o bloque+1 columnas a partir del índice calculado
		MPI_Pack(&x[indice],bloqueEnviado,MPI_DOUBLE,buffer,tamPack,&position,MPI_COMM_WORLD);
		MPI_Pack(&y[indice],bloqueEnviado,MPI_DOUBLE,buffer,tamPack,&position,MPI_COMM_WORLD);

		// Se los mandamos
		MPI_Send(buffer,tamPack,MPI_PACKED,i,0,MPI_COMM_WORLD);
	}
	// Actualizamos el valor de bloque que le corresponde al proceso 0
	if (sobrante > 0) bloque++;
  }
  else {
	// Actualizamos el valor de bloque de cada proceso
	if (nodo < sobrante) bloque++;

	// Reservamos espacio para los vectores
	r = (double *) malloc(sizeof(double)*bloque);
	x = (double *) malloc(sizeof(double)*bloque);
	y = (double *) malloc(sizeof(double)*bloque);

	// Recibimos nuestras porciones de los vectores desde el proceso 0 empaquetadas
	MPI_Recv(buffer,tamPack,MPI_PACKED,0,0,MPI_COMM_WORLD,&status);

	// Desempaquetamos
	position = 0;
	MPI_Unpack(buffer,tamPack,&position,x,bloque,MPI_DOUBLE,MPI_COMM_WORLD);
	MPI_Unpack(buffer,tamPack,&position,y,bloque,MPI_DOUBLE,MPI_COMM_WORLD);
  }
  
  // Calculamos nuestro bloque de P
  #pragma omp parallel for private(i)
  for(i=0;i<bloque;i++)
  {
     r[i]=onetaylor(N,a,x[i],y[i]);
  }

  // El proceso 0 recibe los resultados y los combina en a
  if (nodo == 0) {
	// Retomamos el tamaño inicial de bloque
	if (sobrante > 0) bloque--;

	for (i=1; i<np; i++) {
		// Vemos si le corresponde el tamaño mínimo de bloque o tiene alguna posición extra
		if (i < sobrante) bloqueEnviado = bloque+1;
		else bloqueEnviado = bloque;

		// Calculamos el índice teniendo en cuenta que puede haber bloques de dos tamaños distintos en
		// función de como divida np a P
		if (i<=sobrante) indice = i*(bloque+1);
		else indice = sobrante*(bloque+1) + (i-sobrante)*bloque;

		// Recibimos la r parcial en su posición correspondiente del vector r principal del nodo 0
		MPI_Recv(&r[indice],bloqueEnviado,MPI_DOUBLE,i,0,MPI_COMM_WORLD,&status);
	}
  }
  // El resto envía su vector resultado
  else
	MPI_Send(r,bloque,MPI_DOUBLE,0,0,MPI_COMM_WORLD);

}
